@@ -70,7 +70,7 @@ namespace interop_test
         public void pattern_api_entry()
         {
             // TODO: Why does this not work?
-            Interop.my_api_init_v1(out var api);
+            Interop.pattern_my_api_init_v1(out var api);
             
             var input = new Tupled {x0 = 10};
             var output = api.tupled(input);
